import argparse
import random
from tqdm import tqdm

from util import Graph, Index, PointWithID


def parse_args():
    parser = argparse.ArgumentParser(
        description="Evaluate map inference with geometric metric",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "actual_path", type=str, help="Path to actual/groundtruth .graph file",
    )
    parser.add_argument(
        "inferred_path", type=str, help="Path to inferred/predicted .graph file",
    )
    parser.add_argument(
        "-f",
        "--marker-frequency",
        type=float,
        default=30,
        help="Marker frequency in meters",
    )
    parser.add_argument(
        "-d",
        "--match-distance",
        type=float,
        default=60,
        help="Maximum distance (threshold) in meters to consider a match",
    )

    args = parser.parse_args()
    return args


def get_markers(graph, marker_frequency):
    # we repeatedly select a random unvisited vertex, do a BFS from the vertex,
    # and put down markers while searching
    remaining_vertices = {}
    for vertex in graph.vertices:
        remaining_vertices[vertex.id_] = vertex
    search_queue = []
    markers = []
    pbar = tqdm(total=len(remaining_vertices))

    def add_marker_at(point):
        markers.append(PointWithID(len(markers), point.x, point.y, 0))

    def search_vertex(vertex, remaining_distance):
        if vertex.id_ not in remaining_vertices:
            return
        del remaining_vertices[vertex.id_]
        pbar.update(1)
        for edge in vertex.out_edges:
            if edge.dst.id_ in remaining_vertices:
                search_edge_pos(edge, 0, remaining_distance)

    def search_edge_pos(edge, distance_along_edge, remaining_distance):
        if edge.length() > distance_along_edge + remaining_distance:
            search_queue.append((edge, distance_along_edge + remaining_distance))
            add_marker_at(
                edge.point_along_edge(distance_along_edge + remaining_distance)
            )
        else:
            search_vertex(
                edge.dst, distance_along_edge + remaining_distance - edge.length()
            )

    while len(remaining_vertices) > 0:
        search_queue = []
        vertex = random.choice(list(remaining_vertices.values()))
        if len(vertex.out_edges) == 0 and len(vertex.in_edges) == 0:
            del remaining_vertices[vertex.id_]
            pbar.update(1)
            continue
        add_marker_at(vertex)
        search_vertex(vertex, marker_frequency)
        while len(search_queue) > 0:
            edge, distance_along_edge = search_queue[0]
            search_queue = search_queue[1:]
            search_edge_pos(edge, distance_along_edge, marker_frequency)
    pbar.close()

    return markers


def evaluate(actual_path, inferred_path, marker_frequency, match_distance):
    # read graphs
    actual_graph = Graph.from_file(actual_path)
    inferred_graph = Graph.from_file(inferred_path)

    # get markers
    print("Get markers for {}".format(actual_path))
    actual_markers = get_markers(actual_graph, marker_frequency)
    print("Get markers for {}".format(inferred_path))
    inferred_markers = get_markers(inferred_graph, marker_frequency)

    # match markers between maps
    actual_index = Index(actual_markers)
    actual_matched_set = set()
    inferred_matched_set = set()
    for marker in inferred_markers:
        candidates = actual_index.nearby(marker, match_distance)
        candidates = [
            candidate
            for candidate in candidates
            if marker.distance_to(candidate) < match_distance
        ]
        if candidates:
            for candidate in candidates:
                actual_matched_set.add(candidate.id_)
            inferred_matched_set.add(marker.id_)

    # calculate evaluation metrics
    precision = len(inferred_matched_set) / len(inferred_markers)
    recall = len(actual_matched_set) / len(actual_markers)
    if precision + recall > 0:
        score = 2 * precision * recall / (precision + recall)
    else:
        score = 0
    return precision, recall, score


def main():
    args = parse_args()

    precision, recall, score = evaluate(
        args.actual_path, args.inferred_path, args.marker_frequency, args.match_distance
    )
    print("precision={}, recall={}, score={}".format(precision, recall, score))


if __name__ == "__main__":
    main()
