import argparse
import math
import os
from itertools import tee

import cv2
import numpy as np
from skimage.draw import line as skimage_line
from tqdm import tqdm

from infer_kde_lib import TripLoader


def parse_args():
    parser = argparse.ArgumentParser(
        description="Infer map by KDE",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("trip_dir", type=str, help="input trip directory")
    parser.add_argument("output_dir", type=str, help="output directory")
    parser.add_argument(
        "-c", "--cell-size", type=int, default=2, help="cell size in meters"
    )
    parser.add_argument(
        "-t",
        "--mask-threshold",
        type=int,
        help="mask threshold (turns grayscale into binary)",
    )
    parser.add_argument(
        "-b", "--gaussian-blur", type=int, default=17, help="gaussian blur size"
    )
    parser.add_argument(
        "-s",
        "--voronoi-sampling-interval",
        type=int,
        default=10,
        help="sample one point every so many pixels along the outline",
    )
    parser.add_argument(
        "-d", "--shave-until", type=float, default=0.9999, help="shave until fraction",
    )
    parser.add_argument("-n", "--trip-max", type=int, help="max trips")
    args = parser.parse_args()

    return args


args = parse_args()
all_trips = TripLoader.get_all_trips(args.trip_dir)
if args.mask_threshold is None:
    args.mask_threshold = int(len(all_trips) * 0.05)
if args.trip_max is None:
    args.trip_max = len(all_trips)

cell_size = args.cell_size
mask_threshold = args.mask_threshold
gaussian_blur = args.gaussian_blur
voronoi_sampling_interval = args.voronoi_sampling_interval
MIN_DIR_COUNT = 10
shave_until = args.shave_until
trip_max = args.trip_max

output_dir = args.output_dir
os.makedirs(output_dir, exist_ok=True)

# for cache
output_cache_dir = os.path.join(output_dir, "cache")
use_cache = False
if use_cache:
    os.makedirs(output_cache_dir, exist_ok=True)

# 0 = North, 2 = East, 4 = South, 6 = West
def getsector(fromx, fromy, tox, toy):
    angle = math.atan2(toy - fromy, tox - fromx)
    # subtract pi/8 to align better with cardinal directions
    return int(-angle / (math.pi / 4) + 2) % 8


def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


##
## initialize some globals and read in the trips
##

# all_locations = reduce(lambda x, y: x + y, [t.locations for t in all_trips[:trip_max]])
all_locations = []
for t in all_trips[:trip_max]:
    all_locations += t.locations
all_x = [l.x for l in all_locations]
all_y = [l.y for l in all_locations]

# find bounding box for data
min_x = min(all_x) - 300
max_x = max(all_x) + 300
min_y = min(all_y) - 300
max_y = max(all_y) + 300

diff_x = max_x - min_x
diff_y = max_y - min_y

print(min_x, min_y, max_x, max_y)

width = int(diff_x / cell_size)
height = int(diff_y / cell_size)
xscale = width / diff_x  # pixels per x
yscale = height / diff_y  # pixels per y

# aggregate intensity map for all traces
themap = np.zeros((height, width), dtype=np.uint16)

# aggregate intensity map for all traces, split by sector heading
sector_maps = {}

##
## Build an aggregate intensity map from all the edges
##

filename = os.path.join(output_cache_dir, "n%d_c%d.xml" % (trip_max, cell_size))

if use_cache:
    # TODO cv2 no longer supports XML Load/Save
    pass
    # if os.access(filename, os.F_OK):
    #     print ("Found cached intensity map %s, loading." % (filename))
    #     themap = cv.Load(filename)
    #     for sector in range(8):
    #         sector_maps[sector] = cv.Load(
    #             os.path.join(output_cache_dir, "n%d_c%d_s%d.xml" % (trip_max, cell_size, sector))
    #         )
else:
    # print("Making new intensity map %s." % (filename))
    sector_temp = {}
    for sector in range(8):
        sector_maps[sector] = np.zeros((height, width), dtype=np.uint16)
        sector_temp[sector] = np.zeros((height, width), dtype=np.uint8)

    for trip in all_trips[:trip_max]:
        temp = np.zeros((height, width), dtype=np.uint8)
        for (orig, dest) in pairwise(trip.locations):
            oy = height - int(yscale * (orig.y - min_y))
            ox = int(xscale * (orig.x - min_x))
            dy = height - int(yscale * (dest.y - min_y))
            dx = int(xscale * (dest.x - min_x))
            cv2.line(temp, (ox, oy), (dx, dy), (32), 1, cv2.LINE_AA)

            sector = getsector(ox, oy, dx, dy)
            cv2.line(sector_temp[sector], (ox, oy), (dx, dy), (32), 1, cv2.LINE_AA)

        # accumulate trips into themap
        themap += temp

        # accumulate sector-separated trips into sector_maps
        for sector in range(8):
            sector_maps[sector] += sector_temp[sector]

    # save as XML
    # cv.Save(filename, themap)
    # for sector in range(8):
    #     cv.Save(
    #         os.path.join(output_cache_dir, "n%d_c%d_s%d.xml" % (trip_max, cell_size, sector)),
    #         sector_maps[sector],
    #     )

    lines = np.zeros((height, width), dtype=np.uint8)
    for trip in all_trips:
        for (orig, dest) in pairwise(trip.locations):
            oy = height - int(yscale * (orig.y - min_y))
            ox = int(xscale * (orig.x - min_x))
            dy = height - int(yscale * (dest.y - min_y))
            dx = int(xscale * (dest.x - min_x))
            cv2.line(lines, (ox, oy), (dx, dy), (255), 1, cv2.LINE_AA)
    cv2.imwrite(os.path.join(output_dir, "kde-lines.png"), lines)


print("Intensity map acquired.")


##
## Processing of intensity map below this line
##


def sector_count(line):
    (x0, y0), (x1, y1) = line
    sector = getsector(x0, y0, x1, y1)
    yy, xx = skimage_line(y0, x0, y1, x1)
    li = sector_maps[sector][yy, xx].sum()
    li2 = sector_maps[(sector + 1) % 8][yy, xx].sum()
    li3 = sector_maps[(sector + 7) % 8][yy, xx].sum()
    return li + li2 + li3


def draw_subdiv_facet(img, contours, hierarchy, edge, subdiv):
    lines = []
    t = subdiv.getEdge(edge, 0x13)  # NEXT_AROUND_LEFT = 0x13
    last_t = None
    while t != edge:
        assert t > 4
        if t == last_t:
            break
        last_t = t
        pt = subdiv.edgeOrg(t)[1]
        pt2 = subdiv.edgeDst(t)[1]
        if pt == None or pt2 == None:
            break

        line = (
            (int(round(pt[0])), int(round(pt[1]))),
            (int(round(pt2[0])), int(round(pt2[1]))),
        )

        # cv2.line( img, line[0], line[1], (0,255,0),1,cv2.LINE_AA);

        # if it's inside the top contour, check that it's not in a hole.
        # if it's in a hole, see if there's an inside shape, and so on
        def test_edge(contour_id, adding):
            if contour_id == -1:
                return adding

            contour = contours[contour_id]

            dst1 = cv2.pointPolygonTest(contour, pt, False)
            dst2 = cv2.pointPolygonTest(contour, pt2, False)

            # NOTE improve logic here
            if adding and (dst1 > 0 or dst2 > 0):
                return False
            elif not adding and (dst1 > 0 and dst2 > 0):
                # look at child contour
                return test_edge(hierarchy[0, contour_id, 2], True)
            else:
                # look at next contour
                return test_edge(hierarchy[0, contour_id, 0], adding)

        if pt != pt2 and test_edge(0, False):
            forward_count = sector_count(line)
            reverse_line = (line[1], line[0])
            reverse_count = sector_count(reverse_line)

            # if the count is within a factor of 4, we consider the road bi-directional
            if (
                min(forward_count, reverse_count) > 0
                and max(forward_count, reverse_count)
                / min(forward_count, reverse_count)
                < 4
            ):
                lines.append(line)
                lines.append(reverse_line)
            # if we don't have a count in either direction, the edge is probably a bit crooked.
            # include it as a bi-directional edge - it'll wash out on a long street
            elif forward_count <= MIN_DIR_COUNT and reverse_count <= MIN_DIR_COUNT:
                lines.append(line)
                lines.append(reverse_line)
            elif forward_count > reverse_count:
                lines.append(line)
            else:
                lines.append(reverse_line)

            t = subdiv.getEdge(t, 0x13)  # NEXT_AROUND_LEFT = 0x13

    return lines


def shave_lines(lines):
    nmap = {}
    for line in lines:
        orig = line[0]
        dest = line[1]
        if not orig in nmap:
            nmap[orig] = set()
        if not dest in nmap:
            nmap[dest] = set()

        if orig != dest:
            nmap[orig].add(dest)
            nmap[dest].add(orig)

    newlines = []
    for line in lines:
        if len(nmap[line[0]]) > 1 and len(nmap[line[1]]) > 1:
            newlines.append(line)
    return newlines


def paint_voronoi(contours, hierarchy, img, subdiv):
    # total = len(list(subdiv.edges));
    # print('processing {} edges'.format(total))
    #    elem_size = subdiv.edges.elem_size;

    lines = []

    # counter = 0
    # edges = subdiv.getEdgeList()
    edges = subdiv.getLeadingEdgeList()
    leading_edges_set = set(edges.flatten().tolist())
    other_edges_set = set()
    for edge in leading_edges_set:
        new_edge = edge
        for _ in range(2):
            # add the other two edges of each triangle
            new_edge = subdiv.getEdge(new_edge, 0x13)
            other_edges_set.add(new_edge)
    all_edges_set = leading_edges_set.union(other_edges_set)

    for edge in tqdm(all_edges_set, desc="Processing edges"):
        pt = subdiv.edgeOrg(edge)[1]
        pt2 = subdiv.edgeDst(edge)[1]

        line = (
            (int(round(pt[0])), int(round(pt[1]))),
            (int(round(pt2[0])), int(round(pt2[1]))),
        )

        # cv2.line( img, line[0], line[1], (127,255,0),1,cv2.LINE_AA);

        lines += draw_subdiv_facet(
            img, contours, hierarchy, subdiv.rotateEdge(edge, 1), subdiv,
        )
        lines += draw_subdiv_facet(
            img, contours, hierarchy, subdiv.rotateEdge(edge, 3), subdiv,
        )

    print("Shaving lines")
    while len(lines) > 0:
        oldsize = len(lines)
        lines = shave_lines(lines)
        if len(lines) / float(oldsize) > shave_until:
            break
    print("Done shaving lines")

    vertex_ids = {}
    vertices_by_id = []
    vertex_id = 0
    seen_lines = {}
    edges = []
    # TODO use Graph.write()? Need to take care of cv2.line.
    with open(os.path.join(output_dir, "kde-inferred.graph"), "w") as f:
        for line in lines:
            if line in seen_lines:
                continue
            else:
                seen_lines[line] = 1

            cv2.line(img, line[0], line[1], (255, 0, 0), 1, cv2.LINE_AA)
            if not line[0] in vertex_ids:
                vertex_ids[line[0]] = vertex_id
                vertices_by_id.append(line[0])
                vertex_id += 1
            if not line[1] in vertex_ids:
                vertex_ids[line[1]] = vertex_id
                vertices_by_id.append(line[1])
                vertex_id += 1
            edges.append((vertex_ids[line[0]], vertex_ids[line[1]]))
        for vertex in vertices_by_id:
            f.write(
                "{} {}\n".format(
                    vertex[0] / xscale + min_x, (height - vertex[1]) / yscale + min_y
                )
            )
        f.write("\n")
        for edge in edges:
            f.write("{} {}\n".format(edge[0], edge[1]))


# synthetic data - draw two antialiased lines and add these onto the main map

for sector in range(8):
    sector_maps[sector] = cv2.GaussianBlur(
        sector_maps[sector], (gaussian_blur, gaussian_blur), 0,
    )
    mask = cv2.convertScaleAbs(sector_maps[sector])
    # cv2.imwrite(os.path.join(output_dir, "kde-sector%d.png" % (sector)), mask)
# # create the mask and compute the contour

mask = np.zeros((height, width), dtype=np.uint8)

# histogram creation
# temp = np.zeros((height, width), dtype=np.int32)) # np.uint
# cv.ConvertScale(themap,temp,40,0)
# temp2 = np.zeros((height, width), dtype=np.int32)
# cv.Pow(temp,temp2,0.5)
# cv2.imwrite(os.path.join(output_dir, "kde-histogram.png"),temp2)

themap = cv2.GaussianBlur(themap, (gaussian_blur, gaussian_blur), 0)
# cv2.imwrite(os.path.join(output_dir, "kde-map.png"), themap)
(minval, maxval, minloc, maxloc) = cv2.minMaxLoc(themap)
print("Min: ", minval, " max: ", maxval)
mask = cv2.convertScaleAbs(themap, 255.0 / maxval, 0)  # can also use cv2.normalize?
# cv2.imwrite(os.path.join(output_dir, "kde-mask.png"), mask)
mask = cv2.compare(themap, mask_threshold, cv2.CMP_GT)
cv2.imwrite(os.path.join(output_dir, "kde-thresholded.png"), mask)

contours, hierarchy = cv2.findContours(mask, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_NONE)

# filter out contours with too short perimeters by editing hierarchy to point
# to new neighbors/children/parents as if this contour doesn't exist
removed_contour_ids = []
for contour_id, contour in enumerate(contours):
    if cv2.arcLength(contour, True) < 100:
        next_id, previous_id, first_child_id, parent_id = hierarchy[0, contour_id]
        if previous_id != -1:
            hierarchy[0, previous_id, 0] = next_id
        if next_id != -1:
            hierarchy[0, next_id, 1] = previous_id
        if parent_id != -1 and previous_id == -1:
            # if this contour is the first child of its parent
            hierarchy[0, parent_id, 2] = next_id
        # dealing with the node's descendants looks tricky, but we don't care
        # because we use cv2.RETR_CCOMP so each tree has at most two levels
        removed_contour_ids.append(contour_id)

img = np.zeros((height, width, 3), dtype=np.uint8)

# # create the voronoi graph

subdiv = cv2.Subdiv2D()
subdiv.initDelaunay((0, 0, width, height))


def make_delaunay(contours, img):
    num_points = 0
    for contour_id, contour in enumerate(contours):
        if contour_id in removed_contour_ids:
            continue

        contour = contour.reshape(-1, 2)
        for point in contour[::voronoi_sampling_interval]:
            point = tuple(point)
            # img = cv2.circle(img, point, 3, (0,255,0), 1, cv2.LINE_AA)
            subdiv.insert(point)
            num_points += 1
    print("Inserted %d points to subdiv" % num_points)


make_delaunay(contours, img)

# # draw and save the file
print("Calculating Voronoi")
facetList, facetCenters = subdiv.getVoronoiFacetList([])
print("Done calculating Voronoi")
print("drawing contours")
for contour_id, parent_id in enumerate(hierarchy[0, :, 3]):
    if contour_id in removed_contour_ids:
        continue

    if parent_id == -1:
        color = (255, 255, 255)
    else:
        color = (0, 0, 255)
    img = cv2.drawContours(img, contours, contour_id, color)
print("painting voronoi")
paint_voronoi(contours, hierarchy, img, subdiv)
print("saving image")
cv2.imwrite(os.path.join(output_dir, "kde-voronoi.png"), img)
