import argparse
import math
import os
import random

from tqdm import tqdm

from util import Graph, Index, Point, PointWithID, Trace
from visualize import Visualize


def parse_args():
    parser = argparse.ArgumentParser(
        description="Infer map by K-means",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("trip_dir", type=str, help="input trip directory")
    parser.add_argument("output_dir", type=str, help="output directory")
    parser.add_argument(
        "-s", "--seed-distance", type=float, default=20, help="Seed distance"
    )
    parser.add_argument(
        "-d", "--distance-threshold", type=float, default=70, help="Distance threshold"
    )
    parser.add_argument(
        "-b", "--bearing-threshold", type=float, default=45, help="Bearing threshold",
    )
    parser.add_argument(
        "-m", "--movement-threshold", type=float, default=10, help="Movement threshold",
    )

    args = parser.parse_args()
    return args


class Cluster(Point):
    def __init__(self, center):
        """
        Initialize a cluster.

        center is an instance of Point (see util.py).
        """
        super().__init__(center.x, center.y, center.bearing)
        self.center = center
        self.members = set()

    def add_member(self, point):
        """
        Add a new member to this cluster.

        point is an instance of Point (see util.py).
        """
        self.members.add(point)

    def get_mean(self):
        """
        Returns a Point that is the mean of this cluster's members.

        The x and y positions are simple averages over the members' positions.  The
        bearing is computed by averaging the x and y components of the unit vectors
        represented by the bearing.
        """
        if not self.members:
            return self.center

        x = 0
        y = 0
        bearing_x = 0
        bearing_y = 0
        for member in self.members:
            x += member.x
            y += member.y
            bearing_x += math.cos(math.radians(member.bearing))
            bearing_y += math.sin(math.radians(member.bearing))
        x /= len(self.members)
        y /= len(self.members)
        bearing_x /= len(self.members)
        bearing_y /= len(self.members)
        return Point(x, y, math.degrees(math.atan2(bearing_y, bearing_x)))

    def __eq__(self, other):
        return self.center == other.center and self.members == other.members

    def __repr__(self):
        return "Cluster at {} ({} members)".format(self.center, len(self.members))


def get_markers(traces, seed_distance):
    """
    Create seed points from the provided traces.

    We create seed points every seed_distance meters along the traces, computing the
    bearing as the angle from the positive x axis.

    For example, suppose we have a trace with three observations: (0, 0), (0, 50), and
    (50, 50), and that seed_distance = 30.  We create seed points at (0, 0, 90), (0, 30,
    90), (10, 50, 0), and (40, 50, 0).
    """

    def get_trace_markers(trace):
        """
        Create and return list of seed points for this trace.
        """
        last_distance = seed_distance
        seed_points = []
        for i in range(len(trace.observations) - 1):
            next_obs = trace.observations[i + 1]
            point = trace.observations[i].to_point(next_obs)
            while point.distance_to(next_obs) > seed_distance - last_distance:
                length = point.distance_to(next_obs)
                factor = (seed_distance - last_distance) / length
                point.x, point.y = (
                    (point.x + (next_obs.x - point.x) * factor),
                    (point.y + (next_obs.y - point.y) * factor),
                )
                seed_points.append(Point(point.x, point.y, point.bearing))
                last_distance = 0
            last_distance += point.distance_to(trace.observations[i + 1])
        return seed_points

    # return a list of markers corresponding to each trace
    # we assign global IDs to each marker to use later
    markers_by_trace = []
    next_id = 0
    for trace in traces:
        trace_markers = []
        for marker in get_trace_markers(trace):
            trace_markers.append(
                PointWithID(next_id, marker.x, marker.y, marker.bearing)
            )
            next_id += 1
        markers_by_trace.append(trace_markers)
    return markers_by_trace


def initialize_clusters(markers, distance_threshold, bearing_threshold):
    """
    Create an initial set of clusters.

    Return a list of Cluster objects.

    markers is a list of PointWithID instances.

    We arbitrarily select a marker that has not yet been assigned to a cluster, and
    create a new cluster at the marker.  Then, any markers that are within
    distance_threshold and bearing_threshold of the selected marker are added to the
    cluster.  Note that you have to add the marker that was used as the cluster center
    into the cluster itself.

    For example, suppose that we have five markers.  We randomly select (26, 0, 83) and
    create a new cluster at that point.  The remaining seed points are at (0, 0, 90),
    (30, 0, 90), (50, 10, 0), and (50, 40, 0).  If distance_threshold = 10 and
    bearing_threshold = 10, then (30, 0, 90) is assigned to the existing cluster because
    the distance to the cluster is 4 and the bearing difference is 7, which are both
    within the thresholds.  We repeat the same process until the three remaining markers
    are all assigned to some to-be-created clusters.

    See the Cluster class as defined near the beginning of this file for relevant
    methods.  You might also find Index.nearby (from util.py) useful to query points
    that are within a specified distance to a specified location in 2D.
    """
    # TODO: your code here


def kmeans(markers, initial_clusters, distance_threshold, movement_threshold):
    """
    Run K-means algorithm on the clusters.

    markers is a list of PointWithID instances.  initial_clusters is what you returned
    in initialize_clusters above.

    On each iteration, we first recompute the set of clusters: for every cluster from
    the previous iteration, we initialize one cluster at the mean of the members of the
    old cluster.  Then, we re-assign marker points to the new clusters.

    To re-assign points, distance_threshold is an upper bound for the distance that a
    marker can be from a cluster.  You can use distance_threshold to search for nearby
    clusters based on cluster_index.  However, you should use similarity_metric to
    determine the most similar cluster.

    We continue until the sum of the distance that clusters move on the recomputation
    step is less than movement_threshold.  For example, if we have two clusters (0, 0)
    and (1, 1) with three points (0, 0), (1, 1), and (1.1, 1), then the first point is
    assigned to the first cluster while the other two are assigned to the second
    cluster.  On the next recomputation step, the first cluster moves 0 units, and the
    second cluster moves 0.05 units; if 0 + 0.05 < movement_threshold, then we would
    terminate.
    """

    def similarity_metric(cluster, marker):
        """
        Returns a similarity metric between the specified cluster and marker.

        A lower value indicates greater similarity.
        """
        distance = marker.distance_to(cluster)
        angle_difference = marker.angle_to(cluster)
        return distance + angle_difference

    def recompute_clusters(prev_clusters):
        """
        Returns a tuple of a list of new clusters and the sum of the distances that
        clusters moved.
        """
        # TODO: your code here

    def assign_members(cluster_index, markers):
        """
        Assign each marker point to the closest cluster.
        """
        # TODO: your code here

    clusters = initial_clusters
    distance = float("inf")
    while distance >= movement_threshold:
        clusters, distance = recompute_clusters(clusters)
        cluster_index = Index(clusters)
        assign_members(cluster_index, markers)
        print("moved clusters distance = {}".format(distance))
    return clusters


def generate_edges(graph, markers_by_trace, clusters):
    """
    Connects clusters that appear consecutively in a trace.

    graph is a road network graph where each cluster is a vertex but no edges have been
    added yet.  markers_by_trace is a list, where the i-th element is the list of
    markers for trace i.  clusters is the final clusters returned by kmeans above, but
    each cluster will have an additional vertex field.

    You can use graph.add_edge(src_cluster.vertex, dst_cluster.vertex) to create an
    edge.  If two successive markers belong to different clusters, then we connect those
    clusters.

    You don't need to return anything, just update the graph.
    """

    def get_marker_id_to_cluster_map(clusters):
        """
        Returns a map from marker ID to the cluster that the marker has been assigned to.
        """
        m = {}
        for cluster in clusters:
            for member in cluster.members:
                m[member.id_] = cluster
        return m

    # TODO: your code here


def main():
    args = parse_args()

    # load traces
    traces = Trace.read_traces(args.trip_dir)

    # get initial markers
    print("Getting markers from traces")
    markers_by_trace = get_markers(traces, args.seed_distance)
    flat_markers = [
        marker for trace_markers in markers_by_trace for marker in trace_markers
    ]

    # initialize clusters
    print("Initializing clusters")
    initial_clusters = initialize_clusters(
        flat_markers, args.distance_threshold, args.bearing_threshold
    )

    # run k-means
    print("Running k-means algorithm")
    clusters = kmeans(
        flat_markers,
        initial_clusters,
        2 * args.distance_threshold,
        args.movement_threshold,
    )

    # extract road network
    print("Generating graph")
    graph = Graph()
    for cluster in clusters:
        cluster.vertex = graph.add_vertex(cluster.x, cluster.y)
    generate_edges(graph, markers_by_trace, clusters)

    # output graph
    print("Saving outputs")
    os.makedirs(args.output_dir, exist_ok=True)
    graph.write(os.path.join(args.output_dir, "kmeans-inferred.graph"))

    viz = Visualize()
    viz.draw_graph(graph)
    viz.save(os.path.join(args.output_dir, "kmeans-graph.svg"))

    viz = Visualize()
    viz.draw_points(clusters)
    viz.save(os.path.join(args.output_dir, "kmeans-clusters.svg"))

    viz = Visualize()
    viz.draw_traces(traces, colors="random")
    viz.save(os.path.join(args.output_dir, "kmeans-traces.svg"))


if __name__ == "__main__":
    main()
