import unittest

from infer_kmeans import Cluster, get_markers, initialize_clusters, kmeans
from util import Trace, Observation, Point, PointWithID, Index, Graph


def get_point_id_to_cluster(clusters):
    point_id_to_cluster = {}
    for cluster in clusters:
        for member in cluster.members:
            point_id_to_cluster[member.id_] = cluster
    return point_id_to_cluster


def create_clusters_by_points_ids(points, clusters_by_point_ids):
    clusters = []
    for point_ids in clusters_by_point_ids:
        # arbitrarily use the first point of each group as the center
        cluster = Cluster(points[point_ids[0]])
        for point_id in point_ids:
            cluster.add_member(points[point_id])
        clusters.append(cluster)
    return clusters


class InferKmeansTests(unittest.TestCase):
    def test_get_markers(self):
        # this test simply runs through the example case in the docstring
        trace = Trace([Observation(0, 0), Observation(0, 50), Observation(50, 50)])
        markers = get_markers([trace], 30)[0]

        expected_markers = [
            Point(0, 0, 90),
            Point(0, 30, 90),
            Point(10, 50, 0),
            Point(40, 50, 0),
        ]

        self.assertCountEqual(markers, expected_markers, "Not getting expected markers")

    def test_initialize_clusters(self):
        points = [
            PointWithID(0, 0, 0, 30),  # cluster 1
            PointWithID(1, 2, 2, 35),
            PointWithID(2, 2, 0, 90),  # cluster 2
            PointWithID(3, 1, 1, 85),
            PointWithID(4, 2, 2, 80),
            PointWithID(5, 10, 0, 30),  # cluster 3
            PointWithID(6, 12, 2, 40),
            PointWithID(7, 20, 20, 20),  # cluster 4
        ]
        clusters = initialize_clusters(points, 5, 15)

        expected_clusters_by_point_ids = [[0, 1], [2, 3, 4], [5, 6], [7]]
        expected_clusters = create_clusters_by_points_ids(
            points, expected_clusters_by_point_ids
        )
        self.assert_all_clusters_members_equal(points, clusters, expected_clusters)

    def test_kmeans(self):
        positions = [
            [0, 5],  # cluster 1
            [0, 4],
            [0, 3],
            [1.5, 2],  # cluster 2
            [1.5, 1],
            [2, 0],
            [3, 0],
            [3.5, 1],
            [3.5, 2],
            [5, 3],  # cluster 3
            [5, 4],
            [5, 5],
        ]
        initial_means = [
            [2, 6],
            [2, 3],
            [4.5, 3],
        ]
        points = [
            PointWithID(i, positions[i][0], positions[i][1], 0)
            for i in range(len(positions))
        ]
        initial_clusters = [
            Cluster(Point(mean[0], mean[1], 0)) for mean in initial_means
        ]
        initial_clusters[0].add_member(points[0])
        initial_clusters[1].add_member(points[1])
        initial_clusters[1].add_member(points[2])
        initial_clusters[1].add_member(points[3])
        initial_clusters[1].add_member(points[4])
        initial_clusters[1].add_member(points[5])
        initial_clusters[1].add_member(points[6])
        initial_clusters[2].add_member(points[7])
        initial_clusters[2].add_member(points[8])
        initial_clusters[2].add_member(points[9])
        initial_clusters[2].add_member(points[10])
        initial_clusters[2].add_member(points[11])
        clusters = kmeans(points, initial_clusters, 20, 0.01)

        expected_clusters_by_point_ids = [[0, 1, 2], [3, 4, 5, 6, 7, 8], [9, 10, 11]]
        expected_clusters = create_clusters_by_points_ids(
            points, expected_clusters_by_point_ids
        )
        # ideally, should also check cluster centers
        self.assert_all_clusters_members_equal(points, clusters, expected_clusters)

    def assert_all_clusters_members_equal(self, points, clusters, expected_clusters):
        # check for equality of cluster members, but not cluster centers
        point_id_to_cluster = get_point_id_to_cluster(clusters)
        point_id_to_expected_cluster = get_point_id_to_cluster(expected_clusters)
        for point in points:
            self.assertTrue(
                point.id_ in point_id_to_cluster,
                "{} is not in any cluster".format(point),
            )

            cluster = point_id_to_cluster[point.id_]
            expected_cluster = point_id_to_expected_cluster[point.id_]

            self.assertSetEqual(
                cluster.members,
                expected_cluster.members,
                "Cluster members for {} are not as expected".format(point),
            )


if __name__ == "__main__":
    unittest.main()
