import math
import os

import pyqtree


def float_equals(a, b, epsilon=0.00001):
    return abs(a - b) < epsilon


class Observation(object):
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def to_point(self, next_obs):
        """
        Converts this Observation into a Point.

        A Point has a bearing whereas an Observation does not.  The bearing is
        estimated from next_obs, which should be the observation immediately
        following this one in the parent trace.
        """
        bearing = math.atan2(next_obs.y - self.y, next_obs.x - self.x)
        return Point(self.x, self.y, math.degrees(bearing))


class Point(object):
    def __init__(self, x, y, bearing):
        self.x = float(x)
        self.y = float(y)
        self.bearing = float(bearing)

    def distance_to(self, other):
        """
        Returns the distance to another Point or Observation.
        """
        return math.sqrt(
            (other.x - self.x) * (other.x - self.x)
            + (other.y - self.y) * (other.y - self.y)
        )

    def angle_to(self, other):
        """
        Returns the angle difference to another Point.
        """
        d = other.bearing - self.bearing
        return abs((d + 180) % 360 - 180)

    def __eq__(self, other):
        return (
            float_equals(self.x, other.x)
            and float_equals(self.y, other.y)
            and float_equals(self.bearing, other.bearing)
        )

    def __hash__(self):
        return hash((self.x, self.y, self.bearing))

    def __repr__(self):
        return "Point({}, {}, {})".format(self.x, self.y, self.bearing)


class PointWithID(Point):
    def __init__(self, id_, x, y, bearing):
        super().__init__(x, y, bearing)
        self.id_ = id_

    def __eq__(self, other):
        return super.__eq__(self, other) and (
            not hasattr(other, "id_") or self.id_ == other.id_
        )

    def __hash__(self):
        return hash((self.x, self.y, self.bearing))

    def __repr__(self):
        return "PointWithID({}, {}, {}, id {})".format(
            self.x, self.y, self.bearing, self.id_
        )


class Rectangle(object):
    def __init__(self, min_point, max_point):
        self.min_point = min_point
        self.max_point = max_point

    def lengths(self):
        return self.max_point.x - self.min_point.x, self.max_point.y - self.min_point.y

    def extend_to_contain(self, point):
        self.min_point.x = min(self.min_point.x, point.x)
        self.min_point.y = min(self.min_point.y, point.y)
        self.max_point.x = max(self.max_point.x, point.x)
        self.max_point.y = max(self.max_point.y, point.y)

    def extend_to_contain_rect(self, rect):
        self.extend_to_contain(rect.min_point)
        self.extend_to_contain(rect.max_point)

    @classmethod
    def empty(cls):
        return cls(
            Point(float("inf"), float("inf"), 0), Point(float("-inf"), float("-inf"), 0)
        )


def get_empty_rectangle():
    return Rectangle(
        Point(float("inf"), float("inf"), 0), Point(float("-inf"), float("-inf"), 0)
    )


class Trace(object):
    def __init__(self, observations):
        self.observations = observations

    def bounds(self):
        r = Rectangle.empty()
        for obs in self.observations:
            r.extend_to_contain(obs)
        return r

    @classmethod
    def read_traces(cls, traces_dir):
        files = [os.path.join(traces_dir, fbase) for fbase in os.listdir(traces_dir)]
        files = [fname for fname in files if os.path.isfile(fname)]
        traces = []
        for fname in files:
            with open(fname, "r") as f:
                observations = []
                for line in f:
                    parts = line.strip().split(" ")
                    observations.append(Observation(float(parts[0]), float(parts[1])))
                traces.append(Trace(observations))
        return traces


class Index(object):
    def __init__(self, points):
        """
        Create an index over the specified points.

        Each element must have x and y attributes.
        """
        # compute bounding box
        r = Rectangle.empty()
        for point in points:
            r.extend_to_contain(point)
        self.index = pyqtree.Index(
            bbox=(
                r.min_point.x - 1,
                r.min_point.y - 1,
                r.max_point.x + 1,
                r.max_point.y + 1,
            )
        )

        # insert points into index, with bounding box being unit square around the point
        for point in points:
            self.index.insert(
                point, (point.x - 0.5, point.y - 0.5, point.x + 0.5, point.y + 0.5)
            )

    def nearby(self, point, distance):
        """
        Returns points in the index that are within distance to the specified point.
        """
        # convert to Point object in case something else is provided (for
        # distance_to function)
        point = Point(point.x, point.y, 0)

        # filter points
        candidates = self.index.intersect(
            (
                point.x - distance,
                point.y - distance,
                point.x + distance,
                point.y + distance,
            )
        )
        return [
            candidate
            for candidate in candidates
            if point.distance_to(candidate) < distance
        ]


class Vertex(object):
    def __init__(self, id_, x, y):
        self.id_ = id_
        self.x = x
        self.y = y
        self.in_edges = []
        self.out_edges = []


class Edge(object):
    def __init__(self, id_, src, dst):
        self.id_ = id_
        self.src = src
        self.dst = dst

    def length(self):
        dx = self.dst.x - self.src.x
        dy = self.dst.y - self.src.y
        return math.sqrt(dx * dx + dy * dy)

    def point_along_edge(self, distance):
        """
        Returns the point along this edge that is distance away from the source vertex.
        """
        factor = distance / self.length()
        return Point(
            self.src.x + factor * (self.dst.x - self.src.x),
            self.src.y + factor * (self.dst.y - self.src.y),
            0,
        )


class Graph(object):
    def __init__(self):
        self.vertices = []
        self.edges = []

    def add_vertex(self, x, y):
        vertex = Vertex(len(self.vertices), x, y)
        self.vertices.append(vertex)
        return vertex

    def add_edge(self, src, dst):
        edge = Edge(len(self.edges), src, dst)
        self.edges.append(edge)
        src.out_edges.append(edge)
        dst.in_edges.append(edge)
        return edge

    def write(self, fname):
        with open(fname, "w") as f:
            for vertex in self.vertices:
                f.write("{} {}\n".format(vertex.x, vertex.y))
            f.write("\n")
            for edge in self.edges:
                f.write("{} {}\n".format(edge.src.id_, edge.dst.id_))

    def bounds(self):
        r = Rectangle.empty()
        for vertex in self.vertices:
            r.extend_to_contain(vertex)
        return r

    @classmethod
    def from_file(cls, fname):
        graph = cls()
        with open(fname, "r") as f:
            is_vertex = True
            for line in f:
                line = line.strip()
                if line == "":
                    is_vertex = False
                    continue

                parts = line.split(" ")
                if is_vertex:
                    graph.add_vertex(float(parts[0]), float(parts[1]))
                else:
                    src_id = int(parts[0])
                    dst_id = int(parts[1])
                    graph.add_edge(graph.vertices[src_id], graph.vertices[dst_id])
        return graph
