import argparse
import os
import random

from collections import namedtuple

import svgwrite

from util import Graph, Point, Rectangle, Trace


# element represents graph, trace, or point
DrawingInfo = namedtuple("DrawingInfo", ["element", "color", "width"])


class Visualize:
    """
    Visualize spatial data (graphs, traces, and points) based on svgwrite.
    """

    def __init__(self):
        self.graphs_info = []
        self.traces_info = []
        self.points_info = []

    def draw_graph(self, graph, color="black", width=3):
        self.graphs_info.append(DrawingInfo(graph, color, width))

    def draw_trace(self, trace, color="black", width=3):
        self.traces_info.append(DrawingInfo(trace, color, width))

    def draw_point(self, point, color="black", width=3):
        self.points_info.append(DrawingInfo(point, color, width))

    def draw_traces(self, traces, colors="random", width=3):
        colors = self._get_colors_list(colors, len(traces))
        for trace, color in zip(traces, colors):
            self.draw_trace(trace, color, width)

    def draw_points(self, points, colors="black", width=3):
        colors = self._get_colors_list(colors, len(points))
        for point, color in zip(points, colors):
            self.draw_point(point, color, width)

    def save(self, fname, max_length=1000):
        assert (
            len(self.graphs_info) > 0
            or len(self.traces_info) > 0
            or len(self.points_info) > 0
        ), "At least one of graphs, traces, and points must be provided"

        # automatically determine scale based on the bounding box
        r = Rectangle.empty()
        for graph_info in self.graphs_info:
            r.extend_to_contain_rect(graph_info.element.bounds())
        for trace_info in self.traces_info:
            r.extend_to_contain_rect(trace_info.element.bounds())
        for point_info in self.points_info:
            r.extend_to_contain(point_info.element)

        l = max(r.lengths()[0], r.lengths()[1])
        scale = max_length / l
        lengths = r.lengths()[0] * scale, r.lengths()[1] * scale
        origin = r.min_point
        drawing = svgwrite.Drawing(fname, (int(lengths[0]) + 1, int(lengths[1]) + 1))

        def convert_coords(point):
            return Point(
                int((point.x - origin.x) * scale),
                int(lengths[1] - (point.y - origin.y) * scale),
                0,
            )

        def drawing_add_line(start, end, **kwargs):
            drawing.add(drawing.line((start.x, start.y), (end.x, end.y), **kwargs))

        for graph_info in self.graphs_info:
            graph, color, width = graph_info
            for edge in graph.edges:
                start = convert_coords(edge.src)
                end = convert_coords(edge.dst)
                drawing_add_line(start, end, stroke=color, stroke_width=width)

        for trace_info in self.traces_info:
            trace, color, width = trace_info
            for i in range(len(trace.observations) - 1):
                start = convert_coords(trace.observations[i])
                end = convert_coords(trace.observations[i + 1])
                drawing_add_line(start, end, stroke=color, stroke_width=width)

        for point_info in self.points_info:
            point, color, width = point_info
            c = convert_coords(point)
            drawing.add(drawing.circle(center=(c.x, c.y), r=width, fill=color))

        drawing.save()

    def _random_color(self):
        return svgwrite.rgb(
            random.randint(0, 255), random.randint(0, 255), random.randint(0, 255),
        )

    def _get_colors_list(self, colors, list_len):
        if type(colors) == list:
            assert len(colors) == list_len
            return colors

        # special keyword
        if colors == "random":
            return [self._random_color() for _ in range(list_len)]

        # single color
        return [colors for _ in range(list_len)]


def visualize_graphs_overlay(inferred_dir):
    actual_graph = Graph.from_file(os.path.join("data", "cambridge.graph"))
    kmeans_graph = Graph.from_file(os.path.join(inferred_dir, "kmeans-inferred.graph"))
    kde_graph = Graph.from_file(os.path.join(inferred_dir, "kde-inferred.graph"))

    viz = Visualize()
    # color can be a word or svgwrite.rgb(r, g, b) where r, g, b are in range 0 to 255
    viz.draw_graph(actual_graph, color="grey", width=3)
    viz.draw_graph(kmeans_graph, color="red", width=1)
    viz.draw_graph(kde_graph, color="green", width=1)
    viz.save(os.path.join(inferred_dir, "graphs-overlay.svg"))


def main():
    visualize_graphs_overlay("output")


if __name__ == "__main__":
    main()
